#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Tests for the execution manager. Only one is feasible; other tests are broken need to be fixed Some Fine Day.
"""

# pylint: disable=C0301, E0401, W0212, W0511, W0621
import json
import logging
import sys
from typing import Dict, List
from unittest.mock import patch

import pytest

from workspaces.capability.enums import CapabilityRequestState, QueueState
from workspaces.capability.schema import (
    Capability,
    CapabilityExecution,
    CapabilityRequest,
    CapabilityTemplate,
    CapabilityVersion,
)
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.execution_manager import ExecutionManager
from workspaces.workflow.schema import WorkflowRequest, WorkflowRequestFile
from workspaces.workflow.services.workflow_info import WorkflowInfo
from workspaces.workflow.services.workflow_service import WorkflowService

from .conftest import mock_workflow_requests, test_message_sender

logger = logging.getLogger("execution_manager")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

RECEIVE_MSG_PATCH = "messaging.router.Router.receive_message"


def reset_execution_mgr(ex_man: ExecutionManager) -> ExecutionManager:
    """
    Refresh this execution manager.

    :param ex_man: the execution manager
    :return: a brand-spanking-new exec manager
    """
    ex_man.queues = {}
    ex_man.execution_pool = []
    return ex_man


@pytest.fixture(scope="function")
def workflow_executing_message(
    mock_workflow_service: WorkflowService,
    mock_workflow_requests: List[WorkflowRequest],
) -> Dict:
    """
    Create a reusable workflow-executing message

    :return: Workflow-executing message
    """
    request = mock_workflow_requests[0]
    return {
        "service": "workflow",
        "routing_key": "workflow",
        "type": "workflow-executing",
        "subject": request.__json__(),
    }


@pytest.fixture(scope="function")
def workflow_complete_message(
    mock_workflow_service: WorkflowService, mock_workflow_requests: List[WorkflowRequest]
) -> Dict:
    """
    Create a reusable workflow-complete message

    :return: Workflow-complete message
    """
    request = mock_workflow_requests[0]
    return {
        "service": "workflow",
        "routing_key": "workflow",
        "type": "workflow-complete",
        "subject": request.__json__(),
    }


@pytest.fixture(scope="function")
def fake_execution() -> CapabilityExecution:
    """
    Creates ersatz CapabilityExecution for use in tests

    :return: fake CapabilityExecution
    """
    template = CapabilityTemplate(filename="PPR.xml", content=str.encode("foo"))
    capability = Capability()
    capability.name = "std_calibration"
    capability.id = 1
    capability.templates = [template]

    request = CapabilityRequest()
    request.id = 1
    request.capability = capability

    version = CapabilityVersion()
    version.version_number = 1
    version.request = request

    fake_exec = CapabilityExecution.from_json(
        {
            "capability_request_id": 1,
            "created_at": "2022-01-04T19:40:59.963133",
            "current_workflow_request_id": None,
            "delivery_url": None,
            "id": 1,
            "type": "CapabilityExecution",
            "updated_at": "2022-01-04T19:40:59.963144",
            "version_number": version.version_number,
            "version": version,
        }
    )

    return fake_exec


def test_complete_execution(
    dummy_execution_manager: ExecutionManager,
    fake_execution: CapabilityExecution,
):
    """
    Tests that create_execution correctly marks an execution as complete and sends the
    corresponding message
    """
    with patch(RECEIVE_MSG_PATCH):
        fake_execution.state = CapabilityRequestState.Submitted
        fake_execution.queue_state = QueueState.Queued.name
        with patch("workspaces.capability.message_architect.CapabilityMessageArchitect.compose_message"):
            updated_execution = dummy_execution_manager._complete_execution(fake_execution)
            assert dummy_execution_manager.router.receive_message.called_with(
                {
                    "service": "capability",
                    "routing_key": "capability",
                    "subject": fake_execution.__json__(),
                    "type": "execution-complete",
                }
            )
            assert updated_execution.queue_state == QueueState.NotQueued.name


def test_fail_execution(dummy_execution_manager: ExecutionManager, fake_execution: CapabilityExecution):
    """
    Tests that create_execution correctly marks an execution as failed and sends the
    corresponding message
    """
    with patch(RECEIVE_MSG_PATCH):
        fake_execution.state = CapabilityRequestState.Created.name
        fake_execution.queue_state = QueueState.Queued.name
        with patch("workspaces.capability.message_architect.CapabilityMessageArchitect.compose_message"):
            updated_execution = dummy_execution_manager._fail_execution(fake_execution)
            assert dummy_execution_manager.router.receive_message.called_with(
                {
                    "service": "capability",
                    "routing_key": "capability",
                    "subject": fake_execution.__json__(),
                    "type": "execution-failed",
                }
            )
            assert updated_execution.queue_state == QueueState.NotQueued.name


def test_queue_execution(
    dummy_execution_manager: ExecutionManager, fake_execution: CapabilityExecution, mock_capability_info: CapabilityInfo
):
    """
    Tests that an execution is properly queued
    """
    dummy_execution_manager.queue_execution(fake_execution)
    queued_exec = mock_capability_info.lookup_execution(fake_execution.id)
    assert queued_exec is not None
    assert queued_exec.queue_state == QueueState.Queued.name

    reset_execution_mgr(dummy_execution_manager)


def test_resume_execution(
    dummy_execution_manager: ExecutionManager, fake_execution: CapabilityExecution, mock_capability_info: CapabilityInfo
):
    """
    Tests that a paused execution is resumed

    :return:
    """
    with patch("workspaces.capability.services.capability_info.CapabilityInfo.save_execution"):
        capability = Capability.from_json(
            {
                "name": "std_calibration",
                "type": "Capability",
            }
        )
        fake_execution.capability = capability
        dummy_execution_manager.queue_execution(fake_execution)
        assert fake_execution.queue_state == QueueState.Queued.name
        dummy_execution_manager.resume_execution(fake_execution)
        assert fake_execution.queue_state == QueueState.Running.name

    reset_execution_mgr(dummy_execution_manager)


def test_submit_workflow_request(
    dummy_execution_manager: ExecutionManager,
    fake_execution: CapabilityExecution,
    mock_workflow_requests: List[WorkflowRequest],
    mock_workflow_info: WorkflowInfo,
):
    """
    Tests that a workflow request is submitted

    :return:
    """
    wf_request = mock_workflow_requests[0]
    wf_req_file = WorkflowRequestFile(
        filename="PPR.xml",
        content=str.encode("I'm a little teapot"),
        workflow_request_id=wf_request.workflow_request_id,
    )

    args = {
        "user_email": "radio_astronomer@scientistsЯus.org",
        "product_locator": "Dang! I -knew- I had that stupid product around here someplace!",
        "files_to_transfer": ", ./",
        "need_project_metadata": "true",
    }
    dummy_execution_manager.submit_workflow_request(
        execution_id=fake_execution.id, workflow_name="std_calibration", workflow_args=args, files=[wf_req_file]
    )
    wf_request = mock_workflow_info.lookup_workflow_request(request_id=wf_request.workflow_request_id)
    assert wf_request is not None
