# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
""" Tests for the workflow service """

# pylint: disable=C0301, E0611, W0611, W0621

import json
import logging
import os
import pathlib
import shutil
import sys
from datetime import datetime
from typing import Any, List

# pylint: disable=E0401, R0201, R0913, W0212
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest
from messaging.messenger import SenderIF
from mock_alchemy.mocking import UnifiedAlchemyMagicMock

from workspaces.capability.services.capability_service import Router
from workspaces.workflow.schema import Workflow, WorkflowRequest, WorkflowRequestFile
from workspaces.workflow.services.interfaces import WorkflowInfoIF
from workspaces.workflow.services.workflow_info import WorkflowInfo
from workspaces.workflow.services.workflow_service import WorkflowService

from .conftest import mock_workflow_service

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

FAKE_SUBMIT = "fake.submit"
"""
Tests for WorkflowService
"""


files = [
    WorkflowRequestFile(
        workflow_request_id=-1,
        filename="null.condor",
        content=b"executable = null\n"
        + b"arguments = {{arguments}}\n"
        + b"error = null.err\n"
        + b"log = condor.log\n\n\n"
        + b"queue",
    ),
    WorkflowRequestFile(
        workflow_request_id=-1,
        filename="null.dag",
        content=b"JOB null null.condor",
    ),
    WorkflowRequestFile(
        workflow_request_id=-1,
        filename="null.sh",
        content=b"#!/bin/sh\n\n" + b"null $*",
    ),
]


@pytest.fixture(scope="module")
def mock_workflow_info() -> WorkflowInfo:
    """
    Pytest fixture providing a mock WorkflowInfo object; mocked object can be used
    like a normal WorkflowInfo instance

    :return: Mocked WorkflowInfo
    """

    null_workflow = Workflow(
        workflow_name="null",
    )

    null_workflow_query = (
        [mock.call.query(Workflow), mock.call.filter_by(name="null")],
        [null_workflow],
    )

    mock_session = UnifiedAlchemyMagicMock(data=[null_workflow_query])
    mock_workflow_info = WorkflowInfo(mock_session)

    with patch(
        "workspaces.workflow.services.workflow_info.WorkflowInfo.lookup_workflow_definition", return_value=null_workflow
    ):
        with patch("workspaces.workflow.services.workflow_info.WorkflowInfo.save_request"):
            yield mock_workflow_info


def add_data_to_db_mock(db_mock: WorkflowInfoIF, data: Any):
    """
    Add a query that returns data to a mocked database access object

    :param db_mock: Mocked data access object (either CapabilityInfo or WorkflowInfo)
    :param query: Query that should return the given data
        EX: session.query(Capability).filter_by(name=capability_name).first() turns into
            [mock.call.query(Capability), mock.call.filter_by(name=capability_name), mock.call.first()]
    :param data: List of data to return from query
    """
    db_mock.session.add(data)


@pytest.fixture(scope="module")
def mock_workflow_requests(
    mock_workflow_service: WorkflowService, mock_workflow_info: WorkflowInfoIF
) -> List[WorkflowRequest]:
    """
    Make us some mock workflow requests for testing.

    :param mock_workflow_service:
    :param mock_workflow_info:
    :return:
    """
    requests = []

    r1 = WorkflowRequest(
        workflow_request_id=-1,
        workflow_name="null",
        argument={"parameters": "-g"},
        state="Ready",
        results_dir="",
        created_at=datetime(1996, 9, 10),
        updated_at=datetime(1996, 9, 10),
    )
    requests.append(r1)

    add_data_to_db_mock(mock_workflow_info, r1)

    r2 = WorkflowRequest(
        workflow_request_id=-2,
        workflow_name="std_calibration",
        argument={
            "product_locator": "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",
            "user_email": "chausman@nrao.edu",
        },
        state="Created",
        results_dir="",
        created_at=datetime(2021, 1, 1),
        updated_at=datetime(2021, 1, 1),
    )
    requests.append(r2)

    add_data_to_db_mock(mock_workflow_info, r2)

    return requests


@pytest.fixture(scope="module")
def test_router() -> Router:
    """
    Pytest fixture providing a Router class for testing

    :return: Test Router
    """
    messenger = MockMessageSender()
    router = Router("test", messenger)
    messenger.destination = router
    return router


class MockMessageSender(SenderIF):
    """Pretends to send a message"""

    destination = None

    def send_message(self, **message):
        """
        Don't send this message, but play it on TV

        :param message: message to send as Dict
        :return:
        """
        print(f"sending message: {message}")
        self.destination.receive_message(json.dumps(message))


@pytest.fixture(scope="module")
def test_message_sender() -> SenderIF:
    """
    Pytest fixture providing a Router class for testing

    :return: Test Router
    """
    messenger = MockMessageSender()
    router = Router("test", messenger)
    messenger.destination = router
    return messenger


def test_do_nothing():
    """
    (make test will croak if there isn't at least one test)
    :return:
    """


@pytest.mark.usefixtures("mock_workflow_info", "mock_workflow_service", "test_router")
class TestWorkflowService:
    """Our test class"""

    requests = []

    def test_create_workflow_request(self, mock_workflow_info: WorkflowInfo):
        """
        Is a workflow request created as expected from workflow info?

        :param mock_workflow_info:
        :return:
        """
        null_workflow = mock_workflow_info.lookup_workflow_definition("null")
        assert null_workflow is not None

        request = mock_workflow_info.create_workflow_request(
            null_workflow,
            json.loads('{"argument":"-g"}'),
        )
        self.requests.append(request)
        assert request in self.requests

    def test_attach_file_to_request(
        self,
        mock_workflow_info: WorkflowInfo,
        mock_workflow_requests: List[WorkflowRequest],
    ):
        """
        Is a file attached to workflow requests as expected?

        :param mock_workflow_info:
        :param mock_workflow_requests:
        :return:
        """
        request = mock_workflow_requests[0]
        test_filename = "test.txt"
        test_content = b"testing file additions"
        file = mock_workflow_info.save_file(request=request, filename=test_filename, content=test_content)
        mock_workflow_info.session.add.assert_called_with(file)
        assert mock_workflow_info.session.flush.call_count == 1

    @patch("pathlib.Path.write_bytes")
    @patch("os.chown")
    @patch("os.chmod")
    @pytest.mark.skip("breaks the CI?")
    def test_prepare_files_for_condor(
        self,
        mock_chmod,
        mock_chown,
        mock_write,
        mock_workflow_service: WorkflowService,
        mock_workflow_requests: List[WorkflowRequest],
    ):
        """
        Make sure files are prepared for Condor.

        :param mock_workflow_service:
        :return:
        """
        with patch("os.mkdir"):
            temp = mock_workflow_service._make_temp_directory(mock_workflow_requests[0])
            mock_workflow_service._prepare_files_for_condor(files, temp)
            assert mock_write.call_count == 3

    @patch("workspaces.workflow.schema.Workflow.render_templates")
    @patch("messaging.messenger.MessageSender.send_message")
    @patch("json.loads")
    @patch("subprocess.Popen")
    @pytest.mark.skip("breaks the CI?")
    def test_render_with_metadata(
        self,
        mock_subprocess,
        mock_loads,
        mock_router,
        mock_render,
        mock_workflow_service: WorkflowService,
        mock_workflow_requests: List[WorkflowRequest],
    ):
        """
        Check that we can render a workflow table as expected.

        :param mock_subprocess:
        :param mock_loads:
        :param mock_router:
        :param mock_render:
        :param mock_workflow_service:
        :param mock_workflow_requests:
        :return:
        """
        definition = Workflow(workflow_name="std_calibration")
        request = mock_workflow_requests[1]
        tempdir = pathlib.Path("/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmp1234abcd")
        mock_workflow_service._render_with_metadata(request, tempdir, definition)

        assert mock_subprocess.call_count == 1
        assert mock_loads.call_count == 1
        assert mock_router.call_count == 0
        assert mock_render.call_count == 1

    @patch("workspaces.workflow.services.workflow_info.WorkflowInfo.save_file")
    @pytest.mark.skip("breaks the CI?")
    def test_determine_usable_files(
        self,
        mock_workflow_requests: List[WorkflowRequest],
        mock_workflow_service: WorkflowService,
    ):
        """
        Check that we're determining usable files correctly.

        :param mock_save:
        :param mock_workflow_requests:
        :param mock_workflow_service:
        :return:
        """
        request = mock_workflow_requests[0]
        usable = mock_workflow_service._determine_usable_files(request, files)
        assert len(usable) == 3

    @pytest.mark.skip("breaks the CI?")
    def test_execute_prepared(self, mock_workflow_service: WorkflowService, tmp_path: pathlib.Path):
        # Create fake folder and DAG file
        fake_folder = tmp_path / "test"
        fake_folder.mkdir()
        fake_dagfile = fake_folder / "test.dag"
        fake_dagfile.touch()

        orig_submit_dag = mock_workflow_service._submit_dag
        mock_workflow_service._submit_dag = MagicMock()

        # Test DAG submission
        mock_workflow_service._execute_prepared(fake_folder, False)
        mock_workflow_service._submit_dag.assert_called_with(fake_folder, fake_dagfile, False)

        # DAG cleanup
        mock_workflow_service._submit_dag = orig_submit_dag
        os.remove(fake_dagfile)

        # Create fake submit file
        fake_jobfile = fake_folder / "test.condor"
        fake_jobfile.touch()

        orig_submit_job = mock_workflow_service._submit_job
        mock_workflow_service._submit_job = MagicMock()

        # Test job submission
        mock_workflow_service._execute_prepared(fake_folder, False)
        mock_workflow_service._submit_job.assert_called_with(fake_folder, fake_jobfile, False)

        # Cleanup
        mock_workflow_service._submit_job = orig_submit_job
        shutil.rmtree(tmp_path)

    @pytest.mark.skip("breaks the CI?")
    def test_submit_job(self, mock_workflow_service: WorkflowService):
        fake_folder = pathlib.Path("fake_folder")
        fake_jobfile = pathlib.Path(FAKE_SUBMIT)
        fake_jobfile.touch()
        with patch("workspaces.workflow.services.workflow_service.subprocess.run") as mock_sp_run:
            log_path = mock_workflow_service._submit_job(fake_folder, fake_jobfile, False)
            mock_sp_run.assert_called_with(
                ["condor_submit", fake_jobfile.name],
                cwd=str(fake_folder.absolute()),
                preexec_fn=mock_workflow_service._switch_to_submituser(False),
            )
            assert log_path.name == "fake.log"

        os.remove(fake_jobfile)

    @pytest.mark.skip("breaks the CI?")
    def test_submit_dag(self, mock_workflow_service: WorkflowService):
        fake_folder = pathlib.Path("fake_folder")
        fake_dagfile = pathlib.Path("fake.dag")
        with patch("workspaces.workflow.services.workflow_service.subprocess.run") as mock_sp_run:
            log_path = mock_workflow_service._submit_dag(fake_folder, fake_dagfile, False)
            mock_sp_run.assert_called_with(
                ["condor_submit_dag", "-f", fake_dagfile.name],
                cwd=str(fake_folder.absolute()),
                preexec_fn=mock_workflow_service._switch_to_submituser(False),
            )
            assert log_path.name == "fake.dag.dagman.log"

    @pytest.mark.skip("breaks the CI?")
    def test_get_jobfile_log(self, tmp_path: pathlib.Path, mock_workflow_service: WorkflowService):
        with open(tmp_path / FAKE_SUBMIT, "w") as fake_jobfile:
            # Test with file that doesn't containe "log = " entry
            log_path = mock_workflow_service._get_job_logfile_name(tmp_path / FAKE_SUBMIT)
            assert log_path.name == "fake.log"
            # Write "log = " entry
            fake_jobfile.write("log = test.log")

        # Test with same file that now contains "log = " entry
        log_path = mock_workflow_service._get_job_logfile_name(tmp_path / FAKE_SUBMIT)
        assert log_path.name == "test.log"

        # Remove tmp_path just in case
        shutil.rmtree(tmp_path)

    @pytest.mark.skip("breaks the CI?")
    def test_get_dagfile_log(self, mock_workflow_service: WorkflowService):
        log_path = mock_workflow_service._get_dag_logfile_name(pathlib.Path("fake.dag"))
        assert log_path.name == "fake.dag.dagman.log"
